#define CK_PTR *
#define CK_DECLARE_FUNCTION(returnType, name) \
    returnType name
#define CK_DECLARE_FUNCTION_POINTER(returnType, name) \
    returnType (* name)
#define CK_CALLBACK_FUNCTION(returnType, name) \
    returnType (* name)
#ifndef NULL_PTR
#define NULL_PTR 0
#endif

#include <string.h>
#include <stdio.h>
#include <dlfcn.h>
#include <stdlib.h>
#include <stdbool.h>

#include "pkcs11.h"

#define MAX_OBJECT_COUNT 1

#define TLS_CRYPT_V2_KEY_LEN 32
#define TLS_CRYPT_V2_IV_LENGTH 16

#define SLOT_LIST_MAX_SIZE 10

struct PKCS_Context
{
    CK_SESSION_HANDLE session;
    CK_FUNCTION_LIST_PTR p11_functions;
    CK_VOID_PTR lib_handle;
};

void logger(unsigned long err, char *msg, int line, const char* file, const char *func)
{
    printf("\n0x%lx : %s : %s : %s : %d \n",err,msg,func,file,line);
}

/*
 * Function : load_pkcs11_functions()
 * Description : Load the PKCS 11 functions from the library got from get_library()
 */
CK_RV load_pkcs11_functions(struct PKCS_Context *ctx, const char *lib_path)
{
    CK_RV err = CKR_OK;
    CK_C_GetFunctionList function_symbol_list = NULL_PTR;

    /* Get handle to the library */
    ctx->lib_handle = dlopen(lib_path, RTLD_NOW);
    if ( !ctx->lib_handle )
    {
        err = CKR_GENERAL_ERROR;
        logger(err,"dlopen() failed",__LINE__,__FILE__,__FUNCTION__);
        goto exit;
    }

    /* Obtain address of a symbols in a shared library */
    function_symbol_list = (CK_C_GetFunctionList) dlsym(ctx->lib_handle,"C_GetFunctionList");

    /* Get pkcs11 function list pointer */
    if(function_symbol_list)
    {
        err = function_symbol_list(&ctx->p11_functions);
        if ( err != CKR_OK )
        {
            logger(err,"function_symbol_list() failed",__LINE__,__FILE__,__FUNCTION__);
            goto exit;
        }
    }

    /* Initialize PKCS 11 function library */
    if(ctx->p11_functions)
    {
        err = ctx->p11_functions->C_Initialize(NULL_PTR);
        if( err != CKR_OK )
        {
            logger(err,"C_Initialize() failed",__LINE__,__FILE__,__FUNCTION__);
            goto exit;
        }
    }

exit:
    return err;
}

int connectToPKCS11Token(struct PKCS_Context *ctx, const char *lib_path, const char *pin) {
    unsigned long res;

    res = load_pkcs11_functions(ctx, lib_path);
    if(res != CKR_OK)
    {
        logger(res,"load_pkcs11_functions() failed",__LINE__,__FILE__,__FUNCTION__);
    }

    CK_SLOT_ID *slot_list = (CK_SLOT_ID_PTR) calloc(SLOT_LIST_MAX_SIZE, sizeof(CK_SLOT_ID));
    CK_ULONG slot_list_size = SLOT_LIST_MAX_SIZE;
    res = ctx->p11_functions->C_GetSlotList(true, slot_list, &slot_list_size);
    if (res != CKR_OK)
    {
        logger(res,"C_GetSlotList() failed",__LINE__,__FILE__,__FUNCTION__);
        return false;
    }
    if (slot_list_size < 1)
    {
        logger(0xe0UL,"No Slots / Tokens present!",__LINE__,__FILE__,__FUNCTION__);
    }

    res = ctx->p11_functions->C_OpenSession(slot_list[0],CKF_SERIAL_SESSION | CKF_RW_SESSION, NULL, NULL, &ctx->session);
    if (res != CKR_OK)
    {
        logger(res,"C_OpenSession() failed",__LINE__,__FILE__,__FUNCTION__);
        return false;
    }

    res = ctx->p11_functions->C_Login(ctx->session, CKU_USER, (CK_UTF8CHAR_PTR) pin, strlen(pin));
    if (res != CKR_OK)
    {
        logger(res,"C_Login() failed",__LINE__,__FILE__,__FUNCTION__);
        return false;
    }

    return true;
}

int create_aes_key(struct PKCS_Context *ctx)
{
    ulong res;
    CK_MECHANISM aesKeyMechanism;
    aesKeyMechanism.mechanism = CKM_AES_KEY_GEN;

    CK_BBOOL true_var = CK_TRUE;
    CK_BYTE subject[] = "ServerAESKey";
    ulong key_type = CKK_AES;
    ulong value_len = TLS_CRYPT_V2_KEY_LEN;
    CK_ATTRIBUTE attr_list[] = {
            {CKA_TOKEN,    &true_var,  sizeof(true_var)},
            {CKA_LABEL,    subject,    sizeof(subject)},
            {CKA_ENCRYPT,  &true_var,  sizeof(true_var)},
            {CKA_DECRYPT,  &true_var,  sizeof(true_var)},
            {CKA_KEY_TYPE, &key_type,  sizeof(key_type)},
            {CKA_VALUE_LEN,&value_len, sizeof(value_len)}
    };
    CK_OBJECT_HANDLE aes_key;
    res = ctx->p11_functions->C_GenerateKey(ctx->session, &aesKeyMechanism, attr_list, sizeof(attr_list) / sizeof(attr_list[0]), &aes_key);
    if (res != CKR_OK)
    {
        logger(res, "C_GenerateKey() failed",__LINE__,__FILE__,__FUNCTION__);
        return false;
    }

    return true;
}

int create_hmac_key(struct PKCS_Context *ctx)
{
    ulong res;
    CK_MECHANISM hmacKeyMechanism;
    hmacKeyMechanism.mechanism = CKM_GENERIC_SECRET_KEY_GEN;

    CK_BBOOL true_var = CK_TRUE;
    CK_BBOOL false_var = CK_FALSE;
    CK_BYTE subject[] = "ServerHMACKey";
    ulong key_type = CKK_GENERIC_SECRET;
    ulong value_len = TLS_CRYPT_V2_KEY_LEN;
    CK_ATTRIBUTE attr_list[] = {
            {CKA_TOKEN,    &true_var,  sizeof(true_var)},
            {CKA_LABEL,    subject,    sizeof(subject)},
            {CKA_ENCRYPT,  &false_var, sizeof(false_var)},
            {CKA_DECRYPT,  &false_var, sizeof(false_var)},
            {CKA_SIGN,     &true_var,  sizeof(true_var)},
            {CKA_VERIFY,   &false_var, sizeof(false_var)},
            {CKA_KEY_TYPE, &key_type,  sizeof(key_type)},
            {CKA_VALUE_LEN,&value_len, sizeof(value_len)}
    };
    CK_OBJECT_HANDLE hmac_key;
    res = ctx->p11_functions->C_GenerateKey(ctx->session, &hmacKeyMechanism, attr_list, sizeof(attr_list) / sizeof(attr_list[0]), &hmac_key);
    if (res != CKR_OK)
    {
        logger(res, "C_GenerateKey() failed",__LINE__,__FILE__,__FUNCTION__);
        return false;
    }

    return true;
}

int find_key(struct PKCS_Context *ctx, CK_OBJECT_HANDLE *key, CK_ATTRIBUTE attr_list[], ulong attr_list_len)
{
    CK_ULONG ulObjectCount;
    ulong res;

    res = ctx->p11_functions->C_FindObjectsInit(ctx->session, attr_list, attr_list_len);
    if (res != CKR_OK)
    {
        logger(res, "C_FindObjectsInit() failed",__LINE__,__FILE__,__FUNCTION__);
        return false;
    }

    res = ctx->p11_functions->C_FindObjects(ctx->session, key, MAX_OBJECT_COUNT, &ulObjectCount);
    if (res != CKR_OK)
    {
        logger(res, "C_FindObjects() failed",__LINE__,__FILE__,__FUNCTION__);
        return false;
    }

    res = ctx->p11_functions->C_FindObjectsFinal(ctx->session);
    if (res != CKR_OK)
    {
        logger(res, "C_FindObjectsFinal() failed",__LINE__,__FILE__,__FUNCTION__);
        return false;
    }

    if (ulObjectCount != 1)
    {
        logger(res, "No key was found",__LINE__,__FILE__,__FUNCTION__);
        return false;
    }

    return true;
}

int find_aes_key(struct PKCS_Context *ctx, CK_OBJECT_HANDLE *key)
{
    CK_BBOOL true_var = CK_TRUE;
    CK_BYTE subject[] = "ServerAESKey";
    ulong key_type = CKK_AES;
    ulong value_len = TLS_CRYPT_V2_KEY_LEN;
    CK_ATTRIBUTE attr_list[] = {
            {CKA_TOKEN,    &true_var,   sizeof(true_var)},
            {CKA_LABEL,    subject,     sizeof(subject)},
            {CKA_ENCRYPT,  &true_var,   sizeof(true_var)},
            {CKA_DECRYPT,  &true_var,   sizeof(true_var)},
            {CKA_KEY_TYPE, &key_type,   sizeof(key_type)},
            {CKA_VALUE_LEN,&value_len,  sizeof(value_len)}
    };
    return find_key(ctx, key, attr_list, sizeof(attr_list) / sizeof(attr_list[0]));
}

int find_hmac_key(struct PKCS_Context *ctx, CK_OBJECT_HANDLE *key)
{
    CK_BBOOL true_var = CK_TRUE;
    CK_BBOOL false_var = CK_FALSE;
    CK_BYTE subject[] = "ServerHMACKey";
    ulong key_type = CKK_GENERIC_SECRET;
    ulong value_len = TLS_CRYPT_V2_KEY_LEN;
    CK_ATTRIBUTE attr_list[] = {
            {CKA_TOKEN,    &true_var,  sizeof(true_var)},
            {CKA_LABEL,    subject,    sizeof(subject)},
            {CKA_ENCRYPT,  &false_var, sizeof(false_var)},
            {CKA_DECRYPT,  &false_var, sizeof(false_var)},
            {CKA_SIGN,     &true_var,  sizeof(true_var)},
            {CKA_VERIFY,   &false_var, sizeof(false_var)},
            {CKA_KEY_TYPE, &key_type,  sizeof(key_type)},
            {CKA_VALUE_LEN,&value_len, sizeof(value_len)}
    };
    return find_key(ctx, key, attr_list, sizeof(attr_list) / sizeof(attr_list[0]));
}

int perform_aes_256_ctr(struct PKCS_Context *ctx, unsigned char *data, unsigned long data_len, unsigned char *out,
                        unsigned long *out_len, unsigned char *iv)
{
    ulong res = CKR_FUNCTION_FAILED;
    CK_OBJECT_HANDLE aes_key;
    if(!find_aes_key(ctx, &aes_key))
    {
        logger(res, "Failed finding AES Key",__LINE__,__FILE__,__FUNCTION__);
        return false;
    }

    unsigned long final_out_len = *out_len;
    CK_MECHANISM aesEncryptMechanism;
    CK_AES_CTR_PARAMS params;
    memcpy(params.cb, iv, TLS_CRYPT_V2_IV_LENGTH);
    params.ulCounterBits = TLS_CRYPT_V2_IV_LENGTH * 8;
    aesEncryptMechanism.mechanism = CKM_AES_CTR;
    aesEncryptMechanism.pParameter = &params;
    aesEncryptMechanism.ulParameterLen = sizeof(params);

    res = ctx->p11_functions->C_EncryptInit(ctx->session, &aesEncryptMechanism, aes_key);
    if (res != CKR_OK)
    {
        logger(res, "Failed init encrypt",__LINE__,__FILE__,__FUNCTION__);
        return false;
    }

    res = ctx->p11_functions->C_EncryptUpdate(ctx->session, (CK_BYTE_PTR) data, data_len, out, out_len);
    if (res != CKR_OK)
    {
        logger(res, "Failed update encrypt",__LINE__,__FILE__,__FUNCTION__);
        return false;
    }

    final_out_len -= *out_len;
    res = ctx->p11_functions->C_EncryptFinal(ctx->session, out + *out_len, &final_out_len);
    if (res != CKR_OK)
    {
        logger(res, "Failed final encrypt",__LINE__,__FILE__,__FUNCTION__);
        return false;
    }
    *out_len += final_out_len;

    return true;
}

int perform_sha_256_hmac(struct PKCS_Context *ctx, unsigned char *data, unsigned long data_len, unsigned char *out, unsigned long *out_len)
{
    ulong res = CKR_FUNCTION_FAILED;
    CK_OBJECT_HANDLE hmac_key;
    if(!find_hmac_key(ctx, &hmac_key))
    {
        logger(res, "Failed finding HMAC Key",__LINE__,__FILE__,__FUNCTION__);
        return false;
    }

    CK_MECHANISM hmacMechanism;
    hmacMechanism.mechanism = CKM_SHA256_HMAC;

    res = ctx->p11_functions->C_SignInit(ctx->session, &hmacMechanism, hmac_key);
    if (res != CKR_OK)
    {
        logger(res, "Failed Init Sign",__LINE__,__FILE__,__FUNCTION__);
        return false;
    }

    res = ctx->p11_functions->C_SignUpdate(ctx->session, (CK_BYTE_PTR) data, data_len);
    if (res != CKR_OK)
    {
        logger(res, "Failed Update Sign",__LINE__,__FILE__,__FUNCTION__);
        return false;
    }

    res = ctx->p11_functions->C_SignFinal(ctx->session, out, out_len);
    if (res != CKR_OK)
    {
        logger(res, "Failed Final Sign",__LINE__,__FILE__,__FUNCTION__);
        return false;
    }

    return true;
}

int delete_aes_key(struct PKCS_Context *ctx)
{
    ulong res = CKR_FUNCTION_FAILED;
    CK_OBJECT_HANDLE aes_key;
    if(!find_aes_key(ctx, &aes_key))
    {
        logger(res, "find_aes_key() failed",__LINE__,__FILE__,__FUNCTION__);
        return false;
    }

    res = ctx->p11_functions->C_DestroyObject(ctx->session, aes_key);
    if (res != CKR_OK)
    {
        logger(res, "C_DestroyObject() failed",__LINE__,__FILE__,__FUNCTION__);
        return false;
    }

    return true;
}

int delete_hmac_key(struct PKCS_Context *ctx)
{
    ulong res = CKR_FUNCTION_FAILED;
    CK_OBJECT_HANDLE hmac_key;
    if(!find_hmac_key(ctx, &hmac_key))
    {
        logger(res, "find_hmac_key() failed",__LINE__,__FILE__,__FUNCTION__);
        return false;
    }

    res = ctx->p11_functions->C_DestroyObject(ctx->session, hmac_key);
    if (res != CKR_OK)
    {
        logger(res, "C_DestroyObject() failed",__LINE__,__FILE__,__FUNCTION__);
        return false;
    }

    return true;
}

int main()
{
    struct PKCS_Context ctx;
    if(!connectToPKCS11Token(&ctx, "/usr/lib/pkcs11/libsofthsm2.so", NULL))
    {
        return 1;
    }

    if(!create_aes_key(&ctx))
    {
        return 1;
    }
    const char *test_message = "Hallo";
    unsigned char buffer[5];
    unsigned long res_length = sizeof(buffer);
    unsigned char buffer_inv[5];
    unsigned char iv[16] = { 0 };
    if(!perform_aes_256_ctr(&ctx, (unsigned char *) test_message, strlen(test_message), buffer, &res_length,
                            iv))
    {
        return 1;
    }
    perform_aes_256_ctr(&ctx, buffer, res_length, buffer_inv, &res_length, iv);
    delete_aes_key(&ctx);


} /* main */

